/* 
 ===========================================================================
 =                                                                         =
 =  (C) Copyright 1991-1994 The Trustees of Indiana University             =
 =                                                                         =
 =  Permission to use, copy, modify, and distribute this program for       =
 =  non-commercial use and without fee is hereby granted, provided that    =
 =  this copyright and permission notice appear on all copies and          =
 =  supporting documentation, the name of Indiana University not be used   =
 =  in advertising or publicity pertaining to distribution of the program  =
 =  without specific prior permission, and notice be given in supporting   =
 =  documentation that copying and distribution is by permission of        =
 =  Indiana University.                                                    =
 =                                                                         =
 =  Indiana University makes no representations about the suitability of   =
 =  this software for any purpose. It is provided "as is" without express  =
 =  or implied warranty.                                                   =
 =                                                                         =
 ===========================================================================
 =                                                                         =
 = File:                                                                   =
 =   IUPOP3_VMS.C                                                          =
 =                                                                         =
 = Synopsis:                                                               =
 =   A collection of VMS system services and callable mail functions       =
 =   for the VMS IUPOP3 server.                                            =
 =                                                                         =
 = Authors:                                                                =
 =   Jacob Levanon & Larry Hughes                                          =
 =   Indiana University                                                    =
 =   University Computing Services, Network Applications                   =
 =                                                                         =
 = Credits:                                                                =
 =   This software is based on the Post Office Protocol version 3,         =
 =   as implemented by the University of California at Berkeley.           =
 =                                                                         =
 ===========================================================================
 =                                                                         =
 =  This file based on version 1.8-1 of the iupop3 server, the             =
 =  last version made made Indiana University.                             =
 =                                                                         =
 =  For differences to this version read changes.txt                       =
 =                                                                         =
 = M. Stenns         Jan 1997                                              =
 =                                                                         =
 ===========================================================================
 = 16-Aug-1996  change of the errno.h behavior for multinet                =
 = 21-Aug-1996  change of valid_vms_user() to check for expired accounts   =
 = Ian Miller 10-FEB-1997                                           IM01   =
 = include change by karol zielonko (Digital UCX person)                   =
 = zielonko@ucx.lkg.dec.com to ignore expired passwords.                   =
 = This is conditional on IGNORE_EXPIRED_PASSWORDS.                        =
 = Ian Miller   10-Feb-1997                                         IM02   =
 = establish lib$sig_to_ret as exception handler in mail_purge_waste       =
 = to avoid crashes when this fails.                                       =
 =                                                                         =
 ===========================================================================
*/

#ifdef GNUC
#define variant_union union
#define variant_struct struct
#endif

/* ======================================================================== */
/* Includes */
/* ======================================================================== */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <unixio.h>

#ifndef UCX
#include <sys/ioctl.h>
#else
#if __VMS_VER >= 70100000
#include <ioctl.h>
#endif /* if VMS version >= V7.1 */
#endif /* UCX */
#include <starlet.h>       /* system services */
#include <lib$routines.h>  /* lib$ services */
#include <descrip.h>
#include <ssdef.h>
#include <maildef.h>
#include <uaidef.h>
#include <jpidef.h>
#include <stat.h>
#ifdef MULTINET      /* 16-Aug-1996 */
#include "multinet_root:[multinet.include]errno.h"
#else
#include <errno.h>
#endif /* MULTINET */

#ifdef UCX
#include <in.h>
#else
#include <sys/types.h>
#include <netinet/in.h>
#endif

#ifdef KERBEROS
#include <krb.h>
#include <des.h>
#endif /* KERBEROS */

#include "iupop3_general.h"
#include "iupop3_vms.h"
#include "iupop3_global.h"
#include "version.h"
#include "mailmsgdef.h"

/* ======================================================================== */
/* Defines */
/* ======================================================================== */
#ifdef MULTINET
#define ioctl socket_ioctl
#define ERRNO socket_errno
#else
#define ERRNO errno
#endif


#ifdef IGNORE_EXPIRED_PASSWORDS 
#define CHECK_UAI_FLAGS uai_flags.uai$v_disacnt || uai_flags.uai$v_nomail
#else
#define CHECK_UAI_FLAGS uai_flags.uai$v_disacnt || uai_flags.uai$v_nomail \
           || uai_flags.uai$v_pwd_expired || uai_flags.uai$v_pwd2_expired
#endif

#define START_HEADER_STRING  "============== RFC 822 Headers ==============="

/* ======================================================================== */
/* Prototypes siehe iupop3_general.h */
/* ======================================================================== */
int	AccountExpired (QuadWord BinTime);
int	get_message_line (POP *p, ITEMLIST *inlist, ITEMLIST *outlist);
int	check_external_file (POP * p, char * name);
char  * replace_tabs (char * string);


/* ======================================================================== */
/* Prototypes of mail$xxx rouines (where are the system header?)
/* ======================================================================== */
#define MAIL_FUNCTION_PROTOTYPE \
    unsigned int *context, ITEMLIST *in_item_list , ITEMLIST *out_item_list

int mail$mailfile_begin (MAIL_FUNCTION_PROTOTYPE);
int mail$mailfile_open (MAIL_FUNCTION_PROTOTYPE);
int mail$mailfile_purge_waste (MAIL_FUNCTION_PROTOTYPE);
int mail$mailfile_close (MAIL_FUNCTION_PROTOTYPE);
int mail$mailfile_end (MAIL_FUNCTION_PROTOTYPE);
int mail$message_begin (MAIL_FUNCTION_PROTOTYPE);
int mail$message_delete (MAIL_FUNCTION_PROTOTYPE);
int mail$message_select (MAIL_FUNCTION_PROTOTYPE);
int mail$message_info (MAIL_FUNCTION_PROTOTYPE);
int mail$message_get (MAIL_FUNCTION_PROTOTYPE);
int mail$message_end (MAIL_FUNCTION_PROTOTYPE);
int mail$user_begin (MAIL_FUNCTION_PROTOTYPE);
int mail$user_get_info (MAIL_FUNCTION_PROTOTYPE);
int mail$user_set_info (MAIL_FUNCTION_PROTOTYPE);
int mail$user_end (MAIL_FUNCTION_PROTOTYPE);
int mail$message_copy (MAIL_FUNCTION_PROTOTYPE);

#if defined  UCX &&  __VMS_VER < 70100000
#include "ucx_ioctl.c"
#endif


/* ======================================================================== */
/* Date - VMS to Unix Format */
/* ======================================================================== */
char *date_vms2unix(char *vmsdate)
{
  int  year, month, day, saveyear, dayofweek;
  long julian, j1, j2, j3;
  char *daynames[] = { "Sun", "Mon", "Tues", "Wed", "Thu", "Fri", "Sat" };
  char time[9];
  char monthname[4];
  static char buffer[32];

  memcpy(buffer,vmsdate+7,4); buffer[4] = '\0';
  saveyear = atoi(buffer);
  year = saveyear;

  memcpy(monthname,vmsdate+3,3); monthname[3] = '\0';
  switch (monthname[0])
  {
    case 'J':
      if (monthname[1] == 'A')
        month = 1;
      else if (monthname[2] == 'N')
        month = 6;
      else
        month = 7;
      break;
    case 'F':
      month = 2;
      break;
    case 'M':
      if (monthname[2] == 'R') month = 3; else month = 5;
      break;
    case 'A':
      if (monthname[1] == 'P') month = 4; else month = 8;
      break;
    case 'S':
      month = 9;
      break;
    case 'O':
      month = 10;
      break;
    case 'N':
      month = 11;
      break;
    default:
      month = 12;
      break;
  }
  monthname[1] = tolower(monthname[1]);
  monthname[2] = tolower(monthname[2]);

  memcpy(buffer,vmsdate,2); buffer[2] = '\0';
  day = atoi(buffer);

  memcpy(time,vmsdate+12,8); time[8] = '\0';

  if (year < 1584)
    buffer[0] = '\0';
  else
  {
    if (month > 2)
      month -= 3;
    else
    {
      month += 9;
      year--;
    }
    j1 = 146097 * (year / 100) / 4;
    j2 = 1461L  * (year % 100) / 4;
    j3 = (153L  * month + 2) / 5 + day + 1721119;
    julian = j1 + j2 + j3;
    dayofweek = (int)(++julian % 7);
    sprintf(buffer,"%-3s, %02d %-3s %02d %-8s", daynames[dayofweek], day,
            monthname, (saveyear-1900), time);
  }
  return(buffer);
}

/* ======================================================================== */
/* Get CPU Utilization */
/* ======================================================================== */
float get_cpu (void)
{
  int status;
  int cpu = 0;
  int length = 0;
  static int pid = 0;

  itemopen(outlist_ptr,outlist);
  itemadd(outlist_ptr,sizeof(cpu),JPI$_CPUTIM,&cpu,&length);
  itemclose(outlist_ptr);
  status = sys$getjpiw(0,&pid,0,outlist,0,0,0);
  if (vms_error(status))
  {
    system_log(LOG_ERROR, "sys$getjpiw: %s", vms_message(status));
    cpu = 0;
  }
  return(((float)cpu) / (100.0));
}

/* ======================================================================== */
/* Mail Close File Context */
/* ======================================================================== */
int mail_close_file_context(POP *p)
{
  int status = SS$_NORMAL;

  if (p->file_context)
  {
     status = mail$mailfile_close(&p->file_context,nullist,nullist);
     if (vms_error(status))
       pop_log(LOG_ERROR, p, "mail$mailfile_close: %s", vms_message(status));
     status = mail$mailfile_end(&p->file_context,nullist,nullist);
     if (vms_error(status))
       pop_log(LOG_ERROR, p, "mail$mailfile_end: %s", vms_message(status));
  }
  return status;
}

/* ======================================================================== */
/* Mail Close Message Context */
/* ======================================================================== */
int mail_close_message_context (POP *p)
{
  int status = SS$_NORMAL;

  if (p->message_context)
  {
    status = mail$message_end(&p->message_context,nullist,nullist);
    if (vms_error(status))
      pop_log(LOG_ERROR, p, "mail$message_end: %s", vms_message(status));
  }
  return status;
}

/* ======================================================================== */
/* Mail Close User Context */
/* ======================================================================== */
int mail_close_user_context (POP *p)
{
  int status = SS$_NORMAL;

  if (p->user_context)
  {
    status = mail$user_end(&p->user_context,nullist,nullist);
    if (vms_error(status))
      pop_log(LOG_ERROR, p, "mail$user_end: %s", vms_message(status));
  }
  return status;
}

/* ======================================================================== */
/* Mail Delete Message */
/* ======================================================================== */
int mail_delete_message (POP *p,int *message_id)
{
  int status = SS$_NORMAL;

  if (!p->newmail_selected)
  {
    status = mail_folder_select (p,p->folder_name);
    if (vms_error(status))
      return status;
   }

  itemopen(inlist_ptr,inlist);
  itemadd(inlist_ptr,sizeof(message_id),MAIL$_MESSAGE_ID,message_id,0);
  itemadd(inlist_ptr,0,MAIL$_NOSIGNAL,0,0);
  itemclose(inlist_ptr);
  status = mail$message_delete(&p->message_context,inlist,nullist);
  if (vms_error(status))
    pop_log(LOG_ERROR, p, "mail$message_delete: %s", vms_message(status));

  return status;
}

/* ======================================================================== */
/* Mail Folder Select */
/* ======================================================================== */
int mail_folder_select (POP *p,char *folder_name)
{
  int length = 0;
  int messages = 0;
  int status;

  p->msg_count = 0;

  itemopen(inlist_ptr,inlist);
  itemadd(inlist_ptr,strlen(folder_name),MAIL$_MESSAGE_FOLDER,folder_name,0);
  itemadd(inlist_ptr,0,MAIL$_NOSIGNAL,0,0);
  itemclose(inlist_ptr);

  itemopen(outlist_ptr,outlist);
  itemadd(outlist_ptr,sizeof(messages),MAIL$_MESSAGE_SELECTED,
          &messages,&length);
  itemadd(outlist_ptr,0,MAIL$_NOSIGNAL,0,0);
  itemclose(outlist_ptr);

  status = mail$message_select(&p->message_context,inlist,outlist);
  if (vms_error(status))
  {
    p->msg_count = 0;
    if (status == MAIL$_NOTEXIST)
      status = SS$_NORMAL;
    else
      pop_log(LOG_ERROR, p, "mail$message_select: %s", vms_message(status));
  }
  else
  {
    /* limit the number of newmail messages per connection */
    p->msg_count        = min(messages, max_messages);
    p->newmail_selected = TRUE;
    p->msgs_deleted     = 0;
  }
  return status;
}

/* ======================================================================== */
/* Mail Message Info */
/* ======================================================================== */
int mail_message_info (POP *p,int *message_id)
{
  int  length = 0, size;
  int  lines  = 0;
  int  i, status;
  char external_id[256];
  stat_t statbuf;
  Message *mp;
  int   tmp, is_external=FALSE;
  short msg_flags;

  mp = &p->mptr[*message_id-1];

  memset(external_id, '\0', sizeof(external_id));

  itemopen (inlist_ptr, inlist);
  itemadd (inlist_ptr,sizeof(message_id),MAIL$_MESSAGE_ID,message_id,0);
  itemadd (inlist_ptr,0,MAIL$_NOSIGNAL,0,0);
  itemclose (inlist_ptr);
  itemopen (outlist_ptr, outlist);
  itemadd (outlist_ptr, sizeof(msg_flags), MAIL$_MESSAGE_RETURN_FLAGS,
           &msg_flags, &length);
  itemadd (outlist_ptr,sizeof(external_id)-1,MAIL$_MESSAGE_EXTID,external_id,
           &length);
  itemadd(outlist_ptr,sizeof(mp->lines),MAIL$_MESSAGE_SIZE,&mp->lines,&length);
  itemadd (outlist_ptr,sizeof mp->bindate,MAIL$_MESSAGE_BINARY_DATE, 
           mp->bindate, &length);
  itemadd (outlist_ptr,0,MAIL$_NOSIGNAL,0,0);
  itemclose (outlist_ptr);
  status = mail$message_info (&p->message_context,inlist,outlist);
  if (vms_error(status))
  {
    pop_log (LOG_ERROR, p, "mail$message_info: %s", vms_message(status));
    return (POP_FAILURE);
  }

  /* Check if message was sent with MAIL/FOREIGN */
  if (msg_flags & MAIL$M_EXTNSTD)
  {
    pop_log(LOG_INFO, p, "foreign message detected");
    mp->foreign_msg = TRUE;
  }

  is_external = *external_id;

#ifndef FAST_SCAN
  if (is_external)
  {                             /* Mail file is external to MAIL.MAI */
      char path[256];

      sprintf(path,"%s%s",p->mail_directory,external_id);
      statbuf.st_size = 0;
      if (stat(path,&statbuf) < 0)      /* problems with external file */
      {
        statbuf.st_size = 0;
        if (stat(path,&statbuf) < 0)   /* possible problems in external file */
        {
           pop_log(LOG_ERROR, p, "stat() on %s failed",path);
        }
      }
      pop_log(LOG_DEBUG, p, "stat %s, size %d", path, statbuf.st_size);
      mp->length += statbuf.st_size;
  }
  if (!vms_error(status))
  {
    int hdr_from = 0, hdr_cc = 0, hdr_subject = 0, hdr_to  = 0;

    /* get the size of all header lines */
    status = SS$_NORMAL;
    itemopen(outlist_ptr,outlist);
    itemadd(outlist_ptr,ITEM_LENGTH,MAIL$_MESSAGE_FROM,
          p->retrieve.buffer,&hdr_from);
    itemadd(outlist_ptr,ITEM_LENGTH,MAIL$_MESSAGE_CC,
          p->retrieve.buffer,&hdr_cc);
    itemadd(outlist_ptr,ITEM_LENGTH,MAIL$_MESSAGE_SUBJECT,
          p->retrieve.buffer,&hdr_subject);
    itemadd(outlist_ptr,ITEM_LENGTH,MAIL$_MESSAGE_TO,
          p->retrieve.buffer,&hdr_to);
    itemadd(outlist_ptr,0,MAIL$_NOSIGNAL,0,0);
    itemclose(outlist_ptr);
    status = SS$_NORMAL;
    status = mail$message_info (&p->message_context,inlist,outlist);
    if (vms_error(status))
    {
        pop_log(LOG_ERROR, p, "mail$message_get in headers loop: %s",
               vms_message(status));
        return(POP_FAILURE);
    }
    mp->length += hdr_from + hdr_cc + hdr_subject + hdr_to;
  }
  if (!is_external) 
  { 
    mp->length += mp->lines * 47; /* estimate value */
  }
  /* Add some bytes for reformatted and new headers.    */
  /* No POP3 clients that we know of require this to be */
  /* exact, so we make it larger than it needs to be.   */
  if (mp->length > 0) mp->length += 150;
#else
  mp->length = mp->lines * 47; /* estimate file size from number of records */
#endif
  return status;
}


/* ======================================================================== */
/* Mail Open File Context */
/* ======================================================================== */
int mail_open_file_context (POP *p)
{
  int status;

  strcpy (p->retrieve.buffer,p->mail_directory);
  strncat (p->retrieve.buffer,"MAIL.MAI", 
           sizeof p->retrieve.buffer - sizeof "MAIL.MAI");
  pop_log(LOG_DEBUG, p, "opening %s",p->retrieve.buffer);
  itemopen(inlist_ptr,inlist);
  itemadd(inlist_ptr,0,MAIL$_NOSIGNAL,0,0);
  itemclose(inlist_ptr);
  status = mail$mailfile_begin(&p->file_context,inlist,nullist);
  if (vms_error(status))
  {
    pop_log(LOG_ERROR, p, "mail$mailfile_begin: %s", vms_message(status));
    if (p->file_context)
    {
      int tempsts;

      tempsts = mail$mailfile_end(&p->file_context,nullist,nullist);
      if (vms_error(tempsts))
        pop_log(LOG_ERROR, p, "mail$mailfile_end: %s", vms_message(tempsts));
    }
  }
  else
  {
    itemopen(inlist_ptr,inlist);
    itemadd(inlist_ptr,strlen(p->retrieve.buffer),MAIL$_MAILFILE_NAME,
            &p->retrieve.buffer,0);
    itemadd(inlist_ptr,0,MAIL$_NOSIGNAL,0,0);
    itemclose(inlist_ptr);
    status = mail$mailfile_open(&p->file_context,inlist,nullist);
    if (vms_error(status))
    {
      int tempsts;

      if (status == MAIL$_OPENIN)
      {
        pop_log(LOG_DEBUG, p, "user's mail file does not exist");
        status = SS$_NORMAL;
        p->has_no_mail_file = TRUE;
      }
      else
        pop_log(LOG_ERROR, p, "mail$mailfile_open: %s", vms_message(status));
      tempsts = mail$mailfile_end(&p->file_context,nullist,nullist);
      if (vms_error(tempsts))
        pop_log(LOG_ERROR, p, "mail$mailfile_end: %s", vms_message(tempsts));
    }
  }

  return status;
}

/* ======================================================================== */
/* Mail Open Message Context */
/* ======================================================================== */
int mail_open_message_context (POP *p)
{
  int status;

  itemopen(inlist_ptr,inlist);
  itemadd(inlist_ptr,sizeof(p->file_context),MAIL$_MESSAGE_FILE_CTX,
          &p->file_context,0);
  itemadd(inlist_ptr,0,MAIL$_NOSIGNAL,0,0);
  itemclose(inlist_ptr);
  status = mail$message_begin(&p->message_context,inlist,nullist);
  if (vms_error(status))
  {
    pop_log(LOG_ERROR, p, "mail$message_begin: %s", vms_message(status));
    if (p->message_context) mail_close_message_context(p);
  }
  return status;
}

/* ======================================================================== */
/* Mail Open User Context */
/* ======================================================================== */
int mail_open_user_context (POP *p)
{
  int status;

  p->user_context = 0;
  status = mail$user_begin(&p->user_context,nullist,nullist);
  if (vms_error(status))
  {
    pop_log(LOG_ERROR, p, "mail$user_begin: %s", vms_message(status));
    if (p->user_context) mail_close_user_context(p);
  }
  return status;
}

/* ======================================================================== */
/*  Mail Retrieve Message */
/* ======================================================================== */
int mail_retrieve_message (POP *p)
{
  int status = SS$_NORMAL;
  int retval = POP_FAILURE;
  
  if (default_message_type == DECNET_MESSAGE)
    p->decnet = TRUE;
  else
    p->decnet = FALSE;

  if (!p->newmail_selected)
  {
    status = mail_folder_select(p,p->folder_name);
  }

  if (vms_error(status))
    pop_log(LOG_ERROR, p, "retrieve_message context: %s", vms_message(status));
  else
  {
   if ((p->retrieve.message_id < 1) || (p->retrieve.message_id > p->msg_count))
     return(pop_msg(p, POP_FAILURE, "Message %d does not exist.", 
            p->retrieve.message_id));
   else
   {

      sys$gettim (p->retrieve.Start);
      if (mail_retrieve_message_headers(p))
      {
        if (mail_retrieve_message_text(p))
          retval = POP_SUCCESS;
      }

    }
  }
  return(retval);
}

/* ======================================================================== */
/* check existance of external mail files  */
/* ======================================================================== */
int check_external_file (POP * p, char * name)
{
char filename[256];
FILE * filptr;
int status = TRUE;

  strcpy (filename, p->mail_directory);
  strcat (filename, name);
  filptr = fopen (filename, "r");
  if (!filptr)
    status = FALSE;
  else
    fclose (filptr);
  return status;
}

/* ======================================================================== */
/* get message line (with little error processing)
/* ======================================================================== */
int get_message_line (POP *p, ITEMLIST *inlist, ITEMLIST *outlist)
{
int status;
char * tmpptr;

  p->retrieve.bufsize = 0;
  if (p->retrieve.ext_file)
  {
    if (fgets(p->retrieve.buffer, sizeof(p->retrieve.buffer)-4,
              p->retrieve.ext_file))
    {
      status = SS$_NORMAL;
      if ((tmpptr = strchr (p->retrieve.buffer,'\n'))) *tmpptr = '\0';
      p->retrieve.bufsize = strlen (p->retrieve.buffer);
    }
    else
    {
      p->retrieve.bufsize = 0;
      status = MAIL$_NOMOREREC;
    }
  }
  else
  {
    status = mail$message_get(&p->message_context,inlist,outlist);
    if (vms_error(status))
    {
      pop_log(LOG_DEBUG, p, "mail$message_get (line): %s", vms_message(status));
      if ((status == MAIL$_NOMOREREC) && (p->retrieve.lines_sent < 10))
      {  /* this is probably through a callable mail bug */
        ITEMLIST inlist[3], outlist[3];
        ITEMLIST *inlist_ptr, *outlist_ptr;
        int itmp;

        p->retrieve.bufsize = 0;
        itemopen (inlist_ptr, inlist);
        itemadd (inlist_ptr,sizeof(p->retrieve.message_id),MAIL$_MESSAGE_ID,
                            &p->retrieve.message_id,0);
        itemadd (inlist_ptr,0,MAIL$_NOSIGNAL,0,0);
        itemclose (inlist_ptr);
        itemopen (outlist_ptr, outlist);
        itemadd (outlist_ptr,ITEM_LENGTH,
                 MAIL$_MESSAGE_EXTID,p->retrieve.buffer,&p->retrieve.bufsize);
        itemadd (outlist_ptr,0,MAIL$_NOSIGNAL,0,0);
        itemclose (outlist_ptr);
        status = mail$message_info (&p->message_context,inlist,outlist);
        if (vms_error(status) || !p->retrieve.bufsize)
        {
          if (vms_error(status))
            pop_log(LOG_ERROR, p, "mail$message_info: %s", vms_message(status));
          else
          {
            pop_log (LOG_DEBUG, p, "not a external message");
            status = MAIL$_NOMOREREC;
          }
        }
        else
        {
          p->retrieve.buffer[p->retrieve.bufsize++] = '\0';
	  itmp = strlen (p->mail_directory);
	  memmove (p->retrieve.buffer + itmp, p->retrieve.buffer, 
	           p->retrieve.bufsize);
	  memcpy (p->retrieve.buffer, p->mail_directory, itmp);
          p->retrieve.ext_file = fopen (p->retrieve.buffer,"r");
          if (!p->retrieve.ext_file)
          {
            pop_log (LOG_DEBUG, p, "failure opening external file %s",
                                   p->retrieve.buffer);
            status = MAIL$_NOMOREREC;
          }
          else
          {
            pop_log (LOG_DEBUG, p, "process external file %s",
                                   p->retrieve.buffer);
            p->retrieve.bufsize = 0; /* for callable mail compability */
            *p->retrieve.buffer = '\0';
            itmp = p->retrieve.lines_sent; /* spool to the right position */
            status = SS$_NORMAL;
            while (itmp--) status = get_message_line (p, NULL, NULL); 
          }
        }
      }
    }
  }
  return status;
}


/* ======================================================================== */
/*  restores position to begin of message body */
/* ======================================================================== */
int restore_position (POP *p)
{
  int i,status;

  /* open message */
  itemopen(inlist_ptr, inlist);
  itemadd(inlist_ptr, sizeof(p->retrieve.message_id), MAIL$_MESSAGE_ID,
          &p->retrieve.message_id, 0);
  itemadd(inlist_ptr, 0, MAIL$_NOSIGNAL, 0, 0);
  itemclose(inlist_ptr);
  status = mail$message_get(&p->message_context, inlist, nullist);
  if (vms_error(status))
    pop_log(LOG_ERROR, p, "mail$message_get (rest. 1): %s", vms_message(status));
  /* scroll through mail11 headers */
  itemopen(inlist_ptr,inlist);
  itemadd(inlist_ptr,0,MAIL$_MESSAGE_CONTINUE,0,0);
  itemadd(inlist_ptr,0,MAIL$_NOSIGNAL,0,0);
  itemclose(inlist_ptr);
  itemopen(outlist_ptr,outlist);
  itemadd(outlist_ptr,ITEM_LENGTH,MAIL$_MESSAGE_RECORD,
           p->retrieve.buffer,&p->retrieve.bufsize);
  itemadd(outlist_ptr,0,MAIL$_NOSIGNAL,0,0);
  itemclose(outlist_ptr);
  for (i=1; i <= MAXHEADERS && !vms_error(status); i++)
  {
    status = get_message_line (p, inlist, outlist);
  }
  if (vms_error(status))
    pop_log(LOG_ERROR, p, "mail$message_get (rest. 2): %s", vms_message(status));
  p->retrieve.bufsize = 0;
  return status;
}

#define DECNET_HEADER_FORMAT "Date: %s\r\n\
From: %s\r\n\
Subject: %s\r\n\
To: %s\r\n\
Cc: %s\r\nX-VMS-From: %s\r\n\
X-POP3-Server: %s IUPOP3 V%s\r\n\
X-POP3-ID: %s.%s.%d\r\n"

#define ACCESSFAIL_MESSAGE_FORMAT "\r\n\
An error has been detected in this VMS mail file !!!\r\n\
It seems that the body of the message is missing on the VMS host\r\n\
Use the return address to contact the original sender...\r\n\r\n\
message generated by IUPOP3 V%s on %s\r\n"


/* ======================================================================== */
/* Mail Retrieve Message Headers */
/* ======================================================================== */
int mail_retrieve_message_headers (POP *p)
{
  int  status;
  int  headers;
  int  length = 0;
  short int record_type = 0;
  char *pointer;
  char message_date[32], date[32];
  char from[ITEM_LENGTH+4], xfrom[ITEM_LENGTH+4], cc[ITEM_LENGTH+4], 
       to[ITEM_LENGTH+4], subject[ITEM_LENGTH+4];
  int from_len = 0, cc_len = 0, to_len = 0, subject_len = 0;
  int access_fail = FALSE;
  char date_str[11];
  char time_str[9];

  p->retrieve.send_octets = 0;

  pop_log(LOG_DEBUG, p, "headers follow");
  if (!allocate_send_buffer (p)) return POP_FAILURE;

  /* Get message date and headers */
  itemopen(inlist_ptr, inlist);
  itemadd(inlist_ptr, sizeof(p->retrieve.message_id), MAIL$_MESSAGE_ID,
          &p->retrieve.message_id, 0);
  itemadd(inlist_ptr, 0, MAIL$_NOSIGNAL, 0, 0);
  itemclose(inlist_ptr);
  itemopen(outlist_ptr, outlist);
  itemadd (outlist_ptr,ITEM_LENGTH,MAIL$_MESSAGE_EXTID,
          p->retrieve.buffer,&p->retrieve.bufsize);
  itemadd (outlist_ptr, sizeof(message_date), MAIL$_MESSAGE_DATE,
              message_date, &length);
  itemadd(outlist_ptr,ITEM_LENGTH,MAIL$_MESSAGE_FROM,from,&from_len);
  itemadd(outlist_ptr,ITEM_LENGTH,MAIL$_MESSAGE_CC,cc,&cc_len);
  itemadd(outlist_ptr,ITEM_LENGTH,MAIL$_MESSAGE_SUBJECT,subject,&subject_len);
  itemadd(outlist_ptr,ITEM_LENGTH,MAIL$_MESSAGE_TO,to,&to_len);
  itemadd(outlist_ptr, 0, MAIL$_NOSIGNAL, 0, 0);
  itemclose(outlist_ptr);
  status = mail$message_info (&p->message_context,inlist,outlist);
  if (vms_error(status))
  {
    pop_log(LOG_ERROR, p, "mail$message_info (head): %s", vms_message(status));
    return FALSE;
  }
  if (p->retrieve.bufsize) /* true if mailfile is external */
  {
    p->retrieve.buffer[p->retrieve.bufsize] = '\0';
    if (!check_external_file (p, p->retrieve.buffer)) 
    {
      p->retrieve.message_lines = 0;
      access_fail = TRUE;
      /* mark such a message new as retrieved */
      if (!strcmp(p->retrieve.command, "retr"))
        memcpy (p->retrieve.command, "top ",4);
    }
  }
  get_time(time_str);
  get_date(date_str);
  message_date[length] = '\0';
  strcpy(date, date_vms2unix(message_date));

  /* Prepare for remaining headers */
  from[from_len] = cc[cc_len] = subject[subject_len] = to[to_len] = '\0';

  /* Process remaining headers */
  replace_tabs (from);
  replace_tabs (lower(cc));
  replace_tabs (subject);
  replace_tabs (lower(to));
  strcpy (xfrom, from);

  status = SS$_NORMAL;
  patch_from_line (p, from); /* sets p->decnet to true or false */
  /* send only decnet headers if there is any problem with the mail file */
  if (access_fail)
    p->decnet = TRUE;
  else
    status = restore_position (p);

  if (master_log_level >= LOG_DEBUG)
  {
    /* Print to log for debugging */
    fprintf(log_file, "   Date: %s\n", date);
    fprintf(log_file, "   From: %s\n", from);
    fprintf(log_file, "   Subject: %s\n", subject);
    fprintf(log_file, "   To: %s\n", to);
    fprintf(log_file, "   Cc: %s\n", cc);
    fprintf(log_file, "   X-VMS-From: %s\n", xfrom);
    fprintf(log_file, "   X-POP3-Server: %s IUPOP3 V%s\n", myhostname, VERSION);
    fprintf(log_file, "   X-POP3-ID: %s.%s.%d\n", 
            date_str, time_str, retrieved_messages);
  }

#ifdef IGNORE_MAIL11_HEADERS
  if (!p->decnet)
  {
    int i, is_header = TRUE;
    int lines = p->mptr[p->retrieve.message_id-1].lines;

    headers = 0;
#ifdef SEND_ORIG_HEADER
     /* the header lines are the lines after the last record containing
        "======= RFC 822 Headers =====" */
    is_header = FALSE;
    for (i = 0; i < lines; i++)
    {
      status = get_message_line (p, inlist, outlist);
      if (vms_error(status))
      {
        pop_log(LOG_ERROR,p,"mail$message_get (line %d): %s",
	                     p->retrieve.lines_sent,vms_message(status));
	/* something is wrong with this message */
	p->retrieve.message_lines = min (p->retrieve.message_lines, 
	                                 p->retrieve.lines_sent);
	p->decnet = TRUE;
        break;
      }
      p->retrieve.lines_sent++;
      if (strstr(p->retrieve.buffer,START_HEADER_STRING))
      {
        headers = 0;
        p->retrieve.send_bufsize = 0;
        is_header = TRUE;
      }
      if (is_header)
      {
        headers++;
        if (headers > 1) sndbuf_cpy (p);
      }
    }
    if (p->retrieve.ext_file)
    {
      fclose (p->retrieve.ext_file);
      p->retrieve.ext_file = NULL;
    }
    restore_position (p);
    /* if no headers found assume they are topdown */
    if (headers || p->decnet)
      is_header = FALSE;
    else
      is_header = TRUE;
#endif /* SEND_ORIG_HEADER */

  /* the header lines are from the top of the message to the first
     blank line */
    while (is_header)
    {
      headers++;
      status = get_message_line (p, inlist, outlist);
      p->retrieve.lines_sent++;
      if (vms_error(status))
      {
        is_header = FALSE;
        pop_log(LOG_ERROR,p,"mail$message_get (head): %s",vms_message(status));
	p->decnet = TRUE;
        headers = 0;
        restore_position (p);
        break;
      }
      else if (!p->retrieve.bufsize)
      {
        if (headers > 1) /* first line is everytime empty */
        {
          char *ptr = p->retrieve.send_buffer + p->retrieve.send_bufsize;
          *ptr++ = '\r'; *ptr++ = '\n';
          p->retrieve.send_bufsize += 2;
          is_header = FALSE;
        }
      }
      else
        is_header = sndbuf_cpy (p);
    }

    p->retrieve.lines_sent = 0;
    if (headers > lines) headers = lines;
    p->retrieve.message_lines = min (p->retrieve.message_lines, 
                                     (lines - headers));
    pop_log(LOG_DEBUG, p,
             "header has %d lines, total lines %d, lines to retrieve %d",  
              headers,lines,p->retrieve.message_lines);
  }
#endif /* IGNORE_MAIL11_HEADERS */


  if (p->decnet)
  {
      /* create RFC 822 compatible headers for decnet mails */
     sprintf (p->retrieve.send_buffer,  DECNET_HEADER_FORMAT, date, from, 
              subject, to, cc, xfrom, myhostname, VERSION, 
	      date_str, time_str, retrieved_messages);
     p->retrieve.send_bufsize = strlen (p->retrieve.send_buffer);
     p->retrieve.bufsize = 0;
     if (access_fail)
     {
       sprintf (p->retrieve.send_buffer + p->retrieve.send_bufsize,
                ACCESSFAIL_MESSAGE_FORMAT, VERSION, myhostname);
       p->retrieve.send_bufsize = strlen (p->retrieve.send_buffer);
     }
  }
  p->retrieve.send_octets += p->retrieve.send_bufsize;
  return TRUE;
}


/* ======================================================================== */
/* Mail Retrieve Message Text */
/* ======================================================================== */
int mail_retrieve_message_text (POP *p)
{
  /*
   * Problem:
   *   It is dangerous to let the network write() block us since we
   *   have ASTs disabled.  It is also dangerous to perform a 
   *   non-blocking $QIO for large messages, as the socket buffers
   *   will overflow causing garbled messages.
   *
   * Solution:
   *   Set the socket for non-blocking I/O for the duration of
   *   this function only.  If the sockets buffers are full,
   *   the network write() will fail with errno==EWOULDBLOCK.
   *   The 'retrieved' structure in the POP structure maintains
   *   the context, so when we return to this function after a
   *   timer AST expires, we can pick up where we left off.
   * 
   */
  static int zero = 0;
  static int one  = 1;
  int  bytes;
  int  status;
  int  save_errno;
  int  retval  = FALSE;
  int  sending = TRUE;
  int  reading = TRUE;
#ifdef MULTINET
  /* extern int socket_errno; */  /* 16-Aug-1996 */
#endif /* MULTINET */
  Message *mp;

  /* If message is /FOREIGN, send an informative message instead */
  mp = &p->mptr[p->retrieve.message_id - 1];
  if (mp->foreign_msg)
  {
    pop_log(LOG_INFO, p, "cannot transfer foreign message");
    netprintf(p, "\r\nNote from IUPOP3 server:\r\n");
    netprintf(p, "You have received a foreign (binary) mail message from the sender shown \r\n");
    netprintf(p, "on the 'From:' line above.  That message is now in your VMS MAIL folder.\r\n\r\n");
    netprintf(p, "To access it, log in to your VMS host and use the VMS MAIL 'EXTRACT'\r\n");
    netprintf(p, "command.  If you have any questions, please contact your VMS system\r\n");
    netprintf(p, "administrator or support person.\r\n%s", ENDMULTLINTRANS);
    p->retrieve.completed = TRUE;
    p->retrieve.blocked   = FALSE;
    return(TRUE);
  }

  if (!allocate_send_buffer (p)) return POP_FAILURE;
  if (!p->retrieve.send_ptr) p->retrieve.send_ptr = p->retrieve.send_buffer;
  if (p->retrieve.lines_sent > p->retrieve.message_lines)
  {
    pop_log (LOG_ERROR, p, 
       "mail_retrieve_message_text called with illegal values");
    p->retrieve.lines_sent = p->retrieve.message_lines = 0;
  }

  itemopen(inlist_ptr, inlist);
  itemadd(inlist_ptr, 0, MAIL$_MESSAGE_CONTINUE, 0, 0);
  itemadd(inlist_ptr, 0, MAIL$_NOSIGNAL, 0, 0);
  itemclose(inlist_ptr);

  itemopen(outlist_ptr, outlist);
  itemadd(outlist_ptr, ITEM_LENGTH, MAIL$_MESSAGE_RECORD,
          p->retrieve.buffer, &p->retrieve.bufsize);
  itemadd(outlist_ptr, 0, MAIL$_NOSIGNAL, 0, 0);
  itemclose(outlist_ptr);

  /* Set the socket for non-blocked I/O */
  ioctl(p->sockfd, FIONBIO, (long *)&one);

  while (sending)
  {
    if (p->retrieve.blocked)
    {
      /* pop_log(LOG_DEBUG, p, "continuing"); */
      p->retrieve.blocked = FALSE;  /* hopefully! */
    }
    else if (!p->retrieve.send_bufsize)
    {
      p->retrieve.tries = 0;
      p->retrieve.send_ptr = p->retrieve.send_buffer;
      reading = TRUE;
      /* fill send buffer */
      while ((p->retrieve.lines_sent < p->retrieve.message_lines) && reading)
      {
        if (!p->retrieve.bufsize) /* else old line not yet processed */
        {
          status = get_message_line (p, inlist, outlist);
          if (vms_error(status))
          {
            pop_log (LOG_ERROR, p, "mail$message_get (line %d of %d): %s", 
              p->retrieve.lines_sent, p->retrieve.message_lines, 
              vms_message(status));
            p->retrieve.lines_sent = p->retrieve.message_lines - 1;
            p->retrieve.bufsize = 0;
            sending = FALSE;
          }
        }
        if (reading = sndbuf_cpy (p)) p->retrieve.lines_sent++;
        if (p->retrieve.lines_sent == p->retrieve.message_lines) 
        {
           sprintf (p->retrieve.send_buffer + p->retrieve.send_bufsize, 
                    "\r\n%s",ENDMULTLINTRANS);
           p->retrieve.send_bufsize += 5;
           pop_log (LOG_DEBUG, p, "write ENDMULTLINTRANS");
        }
      }
      p->retrieve.send_octets += p->retrieve.send_bufsize;
    }
    if (p->retrieve.lines_sent == p->retrieve.message_lines) 
      sending = FALSE;

    /* Perform the non-blocked network write */
    if (p->retrieve.send_bufsize)
      bytes = netwrite(p->sockfd, p->retrieve.send_ptr, 
                                p->retrieve.send_bufsize);
    else
      bytes = 0;
    save_errno = ERRNO;
    if (bytes == p->retrieve.send_bufsize)
    {
      pop_log (LOG_DEBUG, p, "%d of %d bytes send (blocked count %d)",
	       bytes, p->retrieve.send_bufsize, p->retrieve.tries);
      net_bytes_written += bytes;
      net_write_count++;
      p->retrieve.send_bufsize = 0;
    }
    else
    {
      sending = FALSE;
      if (bytes > 0)
      {
         /* Some bytes were written, so slide the buffer for next time */
         pop_log (LOG_DEBUG, p, "%d of %d bytes send (blocked count %d)",
	       bytes, p->retrieve.send_bufsize, p->retrieve.tries);
         p->retrieve.tries = 0;
         p->retrieve.send_bufsize -= bytes;
         p->retrieve.send_ptr += bytes;
         net_bytes_written += bytes;
         net_write_count++;
         save_errno = EWOULDBLOCK;
      }
      switch (save_errno)
      {
        case EWOULDBLOCK:
          blocked_count++;
          p->retrieve.blocked = TRUE;
          retval = TRUE;
          break;

        case EPIPE:
          pop_log(LOG_ERROR, p, "network write error: broken pipe");
          p->connected = FALSE;
	  p->retrieve.send_bufsize = 0;
	  /* do not mark this message as retrieved */
          strcpy (p->retrieve.command, "pipe");
          break;

        default:
          pop_log(LOG_ERROR, p, "network write error: errno = %d", save_errno);
          break;
      }
    }
  }

  if (!p->retrieve.send_bufsize && 
       p->retrieve.lines_sent == p->retrieve.message_lines)
  {
    if (!p->retrieve.message_lines) netprintf(p, "\r\n%s", ENDMULTLINTRANS);
    p->retrieve.completed = TRUE;
    p->retrieve.blocked   = FALSE;
    if(strcmp(p->retrieve.command, "retr") == 0)
    { 
      int size;
      QuadWord End, Diff;
      char date_str[11];
      char time_str[9];

      get_time(time_str);
      get_date(date_str);
      sys$gettim (End);lib$sub_times (End,p->retrieve.Start,Diff);
      pop_log(LOG_THREAD, p, "X-POP3-ID: %s.%s.%d",
              date_str, time_str, retrieved_messages);
      pop_log (LOG_DEBUG, p, 
              "%d bytes transferred in %s",
              p->retrieve.send_octets,time_from_bintime(Diff,0));
      retrieved_messages++;
      /* size = p->mptr[p->retrieve.message_id-1].length; */
      size = p->retrieve.send_octets;
      retrieved_octets += size;
      if (size > maximum_octets) maximum_octets = size;
      if (size < minimum_octets) minimum_octets = size;
    }
    if (p->retrieve.ext_file)
    {
      fclose (p->retrieve.ext_file);
      p->retrieve.ext_file = NULL;
    }
  }

  /* Reset the socket for blocked I/O */
  ioctl(p->sockfd, FIONBIO, (long *)&zero);

  return retval;
}


/* ======================================================================== */
/* Mail User Info */
/* ======================================================================== */
int mail_user_info (POP *p)
{
  int  status;
  int  length = 0;

  if (!p->user_context) mail_open_user_context(p);
  itemopen(inlist_ptr,inlist);
  itemadd(inlist_ptr,strlen(p->user),MAIL$_USER_USERNAME,
          &p->user,0);
  itemadd(inlist_ptr,0,MAIL$_NOSIGNAL,0,0);
  itemclose(inlist_ptr);
  itemopen(outlist_ptr,outlist);
  itemadd(outlist_ptr,sizeof(p->mail_directory)-1,MAIL$_USER_FULL_DIRECTORY,
          p->mail_directory,&length);
  itemadd(outlist_ptr,0,MAIL$_NOSIGNAL,0,0);
  itemclose(outlist_ptr);
  status = mail$user_get_info (&p->user_context,inlist,outlist);
  if (vms_error(status))
  {
    pop_log (LOG_ERROR, p, "mail$user_get_info: %s", vms_message(status));
    p->mail_directory[0] = '\0';
  }
  else
  {
    p->mail_directory[length] = '\0';
  }
  mail_close_user_context(p);
  return status;
}

/* ========================================================================= */
/* Valid VMS User */
/* ========================================================================= */
int valid_vms_user (POP *p)
{
  int  status;
  int  retval = FALSE;
  struct flags uai_flags;
  QuadWord ExpirationTime;
  int  length = 0;
  char upcase_username[32];
  struct   dsc$descriptor username_desc;

  username_desc.dsc$w_length  = strlen (p->user) + 1;
  username_desc.dsc$a_pointer =  p->user;
  strcpy(upcase_username, p->user);
  upper(upcase_username);
  username_desc.dsc$w_length  = strlen(upcase_username);
  itemopen(outlist_ptr, outlist);
  itemadd(outlist_ptr, sizeof(uai_flags), UAI$_FLAGS, &uai_flags, &length);
  itemadd(outlist_ptr, sizeof(QuadWord), UAI$_EXPIRATION, &ExpirationTime, 0);
  itemclose(outlist_ptr);

  status = sys$getuai(0, 0, &username_desc, outlist, 0, 0, 0);
  if (vms_error(status))
  {
    pop_log(LOG_ERROR, p, "sys$getuai: %s", vms_message(status));
    p->connected = FALSE;
  }
  else
  {
    if (!(CHECK_UAI_FLAGS))
    {
      if (AccountExpired (ExpirationTime) )
      { 
         retval = FALSE;
         pop_log(LOG_ERROR, p, 
                  "account \"%s\" has expired!", p->user);
      }
      else
      {
        retval = TRUE;
      }
    }
    else
      pop_log(LOG_ERROR, p, 
        "account \"%s\"is administratively disabled", p->user);
  }
  return(retval);
}

/* ========================================================================== */
/* returns true (1) if account has expired */
/* ========================================================================== */
int AccountExpired (QuadWord BinTime)
{
int HasExpired = 0;
if (BinTime[1])  /* false if no expiration date set */
{
 int i, status;
 QuadWord Current;
 unsigned short int AccountTime[7], CurrentTime[7];

 status = sys$numtim (AccountTime,BinTime);
 status = sys$gettim (Current);
 status = sys$numtim (CurrentTime,Current);
 for (i=0; (i < 7) && !(status = (int)AccountTime[i] - (int)CurrentTime[i]);i++);
 if (status < 0) 
 {
   HasExpired = 1;
 }
}
return HasExpired;
} 


/* ========================================================================== */
/* VMS Message String */
/* ========================================================================== */
char *vms_message(unsigned int code)
{
  static   char  message[256];
  unsigned short length = 0;
  struct   dsc$descriptor message_dx;

  memset(message,0,sizeof message);
  message_dx.dsc$w_length  = sizeof(message);
  message_dx.dsc$a_pointer = message;
  sys$getmsg(code,&length,&message_dx,0xf,0);
  message[length] = '\0';
  return message;
}

/* ======================================================================== */
/* Mail Message Move */
/* ======================================================================== */
int mail_message_move (POP *p,int message_id)
{
  int status = SS$_NORMAL;
     
  if (!p->newmail_selected)
  {
    status = mail_folder_select(p, p->folder_name);
  }
  if (vms_error(status))
    return POP_FAILURE;
  else if (!strcmp (p->folder_name, NEWMAIL_FOLDER))
  {
    strcpy (p->retrieve.buffer,p->mail_directory);
    strncat (p->retrieve.buffer,"MAIL.MAI", 
             sizeof p->retrieve.buffer - sizeof "MAIL.MAI");
    itemopen(inlist_ptr,inlist);
    itemadd(inlist_ptr,sizeof(message_id),MAIL$_MESSAGE_ID,&message_id,0);
    itemadd(inlist_ptr,strlen(p->retrieve.buffer),MAIL$_MESSAGE_DEFAULT_NAME,
            &p->retrieve.buffer,0);
    itemadd(inlist_ptr,strlen(MAIL_FOLDER),MAIL$_MESSAGE_FOLDER,MAIL_FOLDER,0);
    itemadd(inlist_ptr,0,MAIL$_MESSAGE_DELETE,0,0);
    itemadd(inlist_ptr,0,MAIL$_NOSIGNAL,0,0);
    itemclose(inlist_ptr);
    status = mail$message_copy(&p->message_context,inlist,nullist);
    if (vms_error(status))
    {
      pop_log(LOG_ERROR, p, "mail$message_copy: %s", vms_message(status));
      return(POP_FAILURE);
    }
  }
  return POP_SUCCESS;
}

/* ======================================================================== */
/* Mail Mark Deleted */
/* ======================================================================== */
int mail_message_new_count (POP *p)
{
    int status = SS$_NORMAL;

    /*
     *  set new message count only if NEWMAIL folder is selected
     */
    if (strcmp(p->folder_name,NEWMAIL_FOLDER)) return status;
    if (!p->user_context) mail_open_user_context(p);
    itemopen(inlist_ptr,inlist);
    itemadd(inlist_ptr,strlen(p->user),MAIL$_USER_USERNAME, &p->user,0);
    itemadd(inlist_ptr, sizeof(p->msg_count),MAIL$_USER_SET_NEW_MESSAGES,
        &p->msg_count,0);
    itemadd(inlist_ptr,0,MAIL$_NOSIGNAL,0,0);
    itemclose(inlist_ptr);
    status = mail$user_set_info(&p->user_context,inlist,nullist);
    if (vms_error(status))
      pop_log(LOG_ERROR, p, "mail$user_set_info: %s", vms_message(status));
    mail_close_user_context (p);
    return status;
}

/* ======================================================================== */
/* MailFile Purge Waste */
/* ======================================================================== */
int mail_purge_waste (POP *p)
{
    int status = SS$_NORMAL;

    vaxc$establish(lib$sig_to_ret);         /* added 11-Jan-1997 *//*IM02*/

    pop_log(LOG_DEBUG, p, "performing PURGE/RECLAIM on mailbox");
    itemopen(inlist_ptr,inlist);
    itemadd(inlist_ptr,0,MAIL$_MAILFILE_RECLAIM,0,0);
    itemadd(inlist_ptr,0,MAIL$_NOSIGNAL,0,0);
    itemclose(inlist_ptr);
    status = mail$mailfile_purge_waste(&p->file_context,inlist,nullist);
    if (vms_error(status))
    {
      pop_log(LOG_ERROR, p, "mail$mailfile_purge_waste: %s",
              vms_message(status));
      return (POP_FAILURE);
    }
    return (POP_SUCCESS);
}


/* ======================================================================== */
/* Read shared secret for apop command from a file                          */
/* in the user's mail directory                                             */
/* return: 1 if success, else 0                                             */
/* ======================================================================== */
int get_shared_secret (POP *p, char * shared_secret, int len)
{
 FILE * secret_file;
 char *tmp;
 int status = 0;

 *shared_secret = '\0';
 if ((mail_open_user_context(p) == SS$_NORMAL) &&
     (mail_user_info(p) == SS$_NORMAL) &&
     (mail_close_user_context(p) == SS$_NORMAL) )
 {
    strcpy (p->retrieve.buffer,p->mail_directory);
    strncat (p->retrieve.buffer, APOP_FILENAME, 
             sizeof p->retrieve.buffer - sizeof APOP_FILENAME);
    secret_file = fopen (p->retrieve.buffer, "r", "shr=get");
    if (secret_file)
    {
      fgets (shared_secret, len, secret_file);
      fclose (secret_file);
      len = strlen (shared_secret);
      for (tmp = &shared_secret[len-1]; 
           len && isspace(*tmp); len--) *tmp-- = '\0';
      if (len >= MIN_PASSWORD_LENGTH)
        status = 1;
      else
      {
         pop_log(LOG_ERROR, p, 
              "shared secret for user %s has insufficient length", p->user);
        *shared_secret = '\0';
      }
    }
    else
    {
      pop_log(LOG_ERROR, p, 
              "failed to open shared secret file for user %s", p->user);
    }
 }
 else
 {
   pop_log(LOG_ERROR, p, "apop: failed to get maildir for user %s", p->user);
 }
 return status;
}

/* ======================================================================== */
/* converts a binary time into ascii time */
/* a pointer to the ascii string is returned */
/* modes: 0   -  date and time */
/*        1   -  time */
/* ======================================================================== */
char * time_from_bintime (unsigned int bintime[2], int mode)
{
  static char time_string[32];
  int time_len = 0;
  STR_DESCR * descr;

  time_string[0] = '\0';
  descr = Descr (time_string);
  descr->dsc$w_length  = sizeof time_string - 1;
  sys$asctim (&time_len,descr,bintime,mode);
  time_string[time_len] = '\0';

  return time_string;
}

/* ======================================================================== */
/* returns the descriptor address of a zero terminated string. */
/* DESCRIP.H is needed. */
/* ======================================================================== */
STR_DESCR * Descr (char * String)
{
static STR_DESCR Descriptor;

 Descriptor.dsc$w_length  = strlen (String);
 Descriptor.dsc$b_dtype   = DSC$K_DTYPE_T;
 Descriptor.dsc$b_class   = DSC$K_CLASS_S;
 Descriptor.dsc$a_pointer = String;
 return &Descriptor;
}
